@@ -32,6 +32,7 @@ const int RAISE = 1;
 const int LOWER = 0;
 
 di_hash_table *seen_items;
+di_hash_table *notinstallables;
 int last_successful_item = -1;
 
 /* Save default priority, to be able to return to it when we have to lower it */
@@ -40,6 +41,9 @@ int default_priority = 1;
 /* Save priority set by main-menu to detect priority changes from the user */
 int local_priority = -1;
 
+/* Force display of the menu at the current priority */
+int display_menu = 0;
+
 static struct debconfclient *debconf;
 
 static int di_config_package(di_system_package *p,
@@ -47,6 +51,26 @@ static int di_config_package(di_system_package *p,
 
 static void modify_debconf_priority (int raise_or_lower);
 
+static char *debconf_priorities[] =
+  {
+    "low",
+    "medium",
+    "high",
+    "critical"
+  };
+
+#define ARRAY_SIZE(a) (sizeof(a)/sizeof(a[0]))
+static int debconf_to_pri (char *priority) {
+	int i;
+
+	if (priority)
+		for (i = 0; (size_t)i < ARRAY_SIZE(debconf_priorities); ++i)
+			if (0 == strcmp(priority, debconf_priorities[i]))
+				return i;
+
+	return -1;
+}
+
 /*
  * qsort comparison function (sort by menu item values, fall back to
  * lexical sort to resolve ties deterministically).
@@ -72,6 +96,13 @@ static void seen_items_key_destroy (void *key)
 	di_free(s);
 }
 
+static void notinstallables_key_destroy (void *key)
+{
+	di_rstring *s = key;
+	di_free(s->string);
+	di_free(s);
+}
+
 int isdefault(di_system_package *p) {
 	int check;
 
@@ -82,21 +113,28 @@ int isdefault(di_system_package *p) {
 }
 
 /*
- * Menu items with a number equal to 99999 or 99900 (deprecated) are
- * intended only to be shown during package selection in anna, but not
- * in the main menu, so mark them uninstallable.
- * For other packages, run the isinstallble maintainer script and check
+ * Menu items with a number equal to 99999 are intended only to be shown
+ * during package selection in anna, but not in the main menu, so mark
+ * them uninstallable.
+ * For other packages, run the isinstallable maintainer script and check
  * its return code.
  */
 bool isinstallable(di_system_package *p) {
 	int check;
 
-	if ((p->installer_menu_item == 99999) || (p->installer_menu_item == 99900))
+	if (p->installer_menu_item == 99999)
 		return false;
 
 	check = di_system_dpkg_package_control_file_exec(&p->p, "isinstallable", 0, NULL);
 	if (check <= 0)
 		return true;
+
+	/* Add to table listing not installable packages */
+	di_rstring *p_name = di_new0(di_rstring, 1);
+	p_name->string = di_stradup(p->p.key.string, p->p.key.size);
+	p_name->size = p->p.key.size;
+	di_hash_table_insert(notinstallables, p_name, p_name);
+
 	return false;
 }
 
@@ -139,6 +177,13 @@ get_default_menu_item(di_slist *list)
 	di_system_package *p;
 	di_slist_node *node;
 
+	/* Create table listing not installable packages from scratch as
+	 * the isinstallable status can change at any time
+	 */
+	di_hash_table_destroy(notinstallables);
+	notinstallables = di_hash_table_new_full(di_rstring_hash, di_rstring_equal,
+						 notinstallables_key_destroy, NULL);
+
 	/* Traverse the list, return the first menu item that isn't installed */
 	for (node = list->head; node != NULL; node = node->next) {
 		p = node->data;
@@ -146,7 +191,7 @@ get_default_menu_item(di_slist *list)
 		if (!p->installer_menu_item ||
 		    p->p.status == di_package_status_installed ||
 		    !isinstallable(p)) {
-			//di_log(DI_LOG_LEVEL_DEBUG, "not menu item; or not installed");
+			//di_log(DI_LOG_LEVEL_DEBUG, "not a menu item, not installed or not installable");
 			continue;
 		}
 		if (p->installer_menu_item >= NEVERDEFAULT) {
@@ -155,7 +200,7 @@ get_default_menu_item(di_slist *list)
 		}
 		if (p->installer_menu_item < last_successful_item &&
 		    di_hash_table_lookup(seen_items, &p->p.key)) {
-			//di_log(DI_LOG_LEVEL_DEBUG, "menu item is before last_successful_item, and is not new");
+			//di_log(DI_LOG_LEVEL_DEBUG, "menu item (%d) is before last_successful_item (%d), and is not new", p->installer_menu_item, last_successful_item);
 			continue;
 		}
 		/* If menutest says this item should be default, make it so */
@@ -196,6 +241,27 @@ static size_t menu_entry(struct debconfclient *debconf, di_system_package *packa
 	return strlen (buf);
 }
 
+/* Priority at which the menu should be displayed */
+static int menu_priority() {
+	static int default_menu_priority = -1;
+	int menu_prio = -1;
+
+	if (default_menu_priority == -1)
+		default_menu_priority = debconf_to_pri(MENU_PRIORITY);
+
+	if (display_menu)
+		menu_prio = local_priority;
+
+	if (menu_prio < 0 ||
+	    (size_t)menu_prio >= ARRAY_SIZE(debconf_priorities))
+		menu_prio = default_menu_priority;
+
+	//di_log(DI_LOG_LEVEL_INFO, "default: %i; debconf: %i; menu: %i",
+	//	default_menu_priority, local_priority, menu_prio);
+
+	return menu_prio;
+}
+
 /* Displays the main menu via debconf and returns the selected menu item. */
 di_system_package *show_main_menu(di_packages *packages, di_packages_allocator *allocator) {
 	di_system_package **package_array, *p;
@@ -204,7 +270,7 @@ di_system_package *show_main_menu(di_packages *packages, di_packages_allocator *
 	di_system_package *menudefault = NULL, *ret = NULL;
 	int i = 0, num = 0;
 	char buf[256], *menu, *s;
-	int menu_size, menu_used, size;
+	int menu_prio, menu_size, menu_used, size;
 
 	for (node = packages->list.head; node; node = node->next) {
 		p = node->data;
@@ -264,7 +330,8 @@ di_system_package *show_main_menu(di_packages *packages, di_packages_allocator *
 		di_log(DI_LOG_LEVEL_INFO, "no default menu item"); 
 		modify_debconf_priority(LOWER);
 	}
-	debconf_input(debconf, MENU_PRIORITY, MAIN_MENU);
+	menu_prio = menu_priority();
+	debconf_input(debconf, debconf_priorities[menu_prio], MAIN_MENU);
 	debconf_go(debconf);
 	debconf_get(debconf, MAIN_MENU);
 	s = strdup(debconf->value);
@@ -448,48 +515,19 @@ static int do_menu_item(di_system_package *p) {
 	return di_config_package(p, satisfy_virtual);
 }
 
-static char *debconf_priorities[] =
-  {
-    "low",
-    "medium",
-    "high",
-    "critical"
-  };
-
-#define ARRAY_SIZE(a) (sizeof(a)/sizeof(a[0]))
 static void modify_debconf_priority (int raise_or_lower) {
-	int pri, i;
-	static int menu_pri = -1;
+	int pri;
 	const char *template = "debconf/priority";
 	debconf_get(debconf, template);
 	
-	if (menu_pri == -1) {
-		for (i = 0; (size_t)i < ARRAY_SIZE(debconf_priorities); ++i) {
-			if (0 == strcmp(MENU_PRIORITY,
-					debconf_priorities[i]) ) {
-				menu_pri = i;
-				break;
-			}
-		}
-	}
+	pri = debconf_to_pri(debconf->value);
+	if ( pri == -1 )
+		pri = 1;
 
-	pri = 1;
-	if ( debconf->value ) {
-		for (i = 0; (size_t)i < ARRAY_SIZE(debconf_priorities); ++i) {
-			if (0 == strcmp(debconf->value,
-					debconf_priorities[i]) ) {
-				pri = i;
-				break;
-			}
-		}
-	}
 	if (raise_or_lower == LOWER) {
 		--pri;
-		/* Make sure the menu is always displayed after a single
-		 * backing up.
-		 */
-		if (menu_pri != -1 && pri > menu_pri)
-			pri = menu_pri;
+		/* Make sure the menu is always displayed after an error */
+		display_menu = 1;
 	}
 	else if (raise_or_lower == RAISE)
 		++pri;
@@ -509,22 +547,13 @@ static void modify_debconf_priority (int raise_or_lower) {
 }
 	
 static void adjust_default_priority (void) {
-	int pri, i;
+	int pri;
 	const char *template = "debconf/priority";
-	debconf_get(debconf,  template);
+	debconf_get(debconf, template);
 
-	pri = 1;
-	if ( debconf->value ) {
-		for (i = 0; (size_t)i < ARRAY_SIZE(debconf_priorities); ++i) {
-			if (0 == strcmp(debconf->value,
-					debconf_priorities[i]) ) {
-				pri = i;
-				break;
-			}
-		}
-	}
+	pri = debconf_to_pri(debconf->value);
 	
-	if ( pri != local_priority) {
+	if (pri > -1 && pri != local_priority) {
 		if (local_priority > -1)
 			di_log(DI_LOG_LEVEL_INFO, "Priority changed externally, setting main-menu default to '%s' (%s)",
 				debconf_priorities[pri] ? debconf_priorities[pri] : "(null)", debconf->value);
@@ -536,9 +565,12 @@ static void adjust_default_priority (void) {
 static void restore_default_priority (void) {
 	const char *template = "debconf/priority";
 
-	di_log(DI_LOG_LEVEL_INFO, "Restoring default debconf priority '%s'",
-		debconf_priorities[default_priority] ? debconf_priorities[default_priority] : "(null)");
-	debconf_set(debconf, template, debconf_priorities[default_priority]);
+	if (local_priority != default_priority) {
+		di_log(DI_LOG_LEVEL_INFO, "Restoring default debconf priority '%s'",
+			debconf_priorities[default_priority] ? debconf_priorities[default_priority] : "(null)");
+		debconf_set(debconf, template, debconf_priorities[default_priority]);
+		local_priority = default_priority;
+	}
 }
 
 void notify_user_of_failure (di_system_package *p) {
@@ -610,8 +642,7 @@ int main (int argc __attribute__ ((unused)), char **argv) {
 	di_system_package *p;
 	di_packages *packages;
 	di_packages_allocator *allocator;
-	int ret;
-	int last_item_backup = 0;
+	int ret, exit_loop;
 
 	debconf = debconfclient_new();
 	di_system_init(basename(argv[0]));
@@ -625,19 +656,24 @@ int main (int argc __attribute__ ((unused)), char **argv) {
 	    debconf->value && *debconf->value)
 		debconf_set(debconf, template, debconf->value);
 
+	/* Initialize internal priority variables */
+	adjust_default_priority();
+
 	menu_startup();
 
 	seen_items = di_hash_table_new_full(di_rstring_hash, di_rstring_equal,
 					    seen_items_key_destroy, NULL);
+	notinstallables = di_hash_table_new_full(di_rstring_hash, di_rstring_equal,
+						 notinstallables_key_destroy, NULL);
 
+	exit_loop = 0;
 	allocator = di_system_packages_allocator_alloc ();
 	packages = di_system_packages_status_read_file(DI_SYSTEM_DPKG_STATUSFILE, allocator);
-	while ((p=show_main_menu(packages, allocator))) {
+	while (!exit_loop && (p=show_main_menu(packages, allocator))) {
 		di_slist_node *node;
 
-		if (p->installer_menu_item < NEVERDEFAULT && last_item_backup) {
-			restore_default_priority();
-			last_item_backup = 0;
+		if (p->installer_menu_item < NEVERDEFAULT && display_menu) {
+			display_menu = 0;
 		}
 		ret = do_menu_item(p);
 		adjust_default_priority();
@@ -646,21 +682,20 @@ int main (int argc __attribute__ ((unused)), char **argv) {
 				/* Success */
 				if (p->installer_menu_item < NEVERDEFAULT) {
 					last_successful_item = p->installer_menu_item;
-					modify_debconf_priority(RAISE);
+					restore_default_priority();
 					//di_log(DI_LOG_LEVEL_DEBUG, "Installed package '%s', raising last_successful_item to %d", p->p.package, p->installer_menu_item);
 				}
 				else {
-					/* Do not raise the priority after it has been set by cdebconf-priority. */
-					if (0 == strcmp("cdebconf-priority", p->p.package)) {
-						last_item_backup = 0;
-					}
 					// di_log(DI_LOG_LEVEL_DEBUG, "Installed package '%s' but no raise since %d >= %i", p->p.package, p->installer_menu_item, NEVERDEFAULT);
 				}
 				break;
 			case EXIT_BACKUP:
 				di_log(DI_LOG_LEVEL_INFO, "Menu item '%s' succeeded but requested to be left unconfigured.", p->p.package); 
-				modify_debconf_priority(LOWER);
-				last_item_backup = 1;
+				display_menu = 1;
+				break;
+			case EXIT_INSTALLER:
+				/* Interrupt the loop and exit properly */
+				exit_loop = 1;
 				break;
 			default:
 				di_log(DI_LOG_LEVEL_WARNING, "Menu item '%s' failed.", p->p.package);
@@ -675,7 +710,8 @@ int main (int argc __attribute__ ((unused)), char **argv) {
 			seen_name->string = di_stradup(seen->p.key.string,
 						       seen->p.key.size);
 			seen_name->size = seen->p.key.size;
-			di_hash_table_insert(seen_items, seen_name, seen_name);
+			if (! di_hash_table_lookup(notinstallables, &seen->p.key))
+				di_hash_table_insert(seen_items, seen_name, seen_name);
 		}
 
 		di_packages_free (packages);
@@ -684,10 +720,10 @@ int main (int argc __attribute__ ((unused)), char **argv) {
 		packages = di_system_packages_status_read_file(DI_SYSTEM_DPKG_STATUSFILE, allocator);
 
 		/* tell cdebconf to save the database */
-		debconf->command(debconf, "X_SAVE", NULL);
+		debconf_x_save(debconf);
 	}
 	
-	return EXIT_FAILURE;
+	return exit_loop != 0 ? EXIT_OK : EXIT_FAILURE;
 }
 
 int di_config_package_depth=0;
@@ -747,6 +783,7 @@ static int di_config_package(di_system_package *p,
 	free(configcommand);
 	switch (ret) {
 		case EXIT_OK:
+		case EXIT_INSTALLER:
 			p->p.status = di_package_status_installed;
 			break;
 		default:
